#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gtk/gtk.h>

#include "device.h"
#include "callbacks.h" /* for the enum */

/* this device "class" could easily be changed to a structure to allow
 * handling of more than one device instead of running two instances of
 * the program */
static struct ics_interface * iface;
static struct ics_stream_state * ss;
static struct ics_pass_iface_and_stream *tempIfaceAndStream;
static struct ics_pass_list_store_and_integer * tempListStoreAndNumPackets;
static guint numDevStatPackets;
static gint incomingPacketHandlerTag;
static GIOChannel * iochan;

/* callbacks first to eliminate redundant prototypes */
static void
dev_stat_callback(struct ics_stream_state *ss, struct ics_dev_status *ds,
		void *client_data)
{
	numDevStatPackets++;
}

static void
rx_overflow_callback(struct ics_stream_state *ss, struct ics_rx_overflow *rxo,
		void *user_data)
{
	GtkListStore * listStore = (GtkListStore *) user_data;
	GtkTreeIter iter;

	gtk_list_store_append(listStore, &iter);
	gtk_list_store_set(listStore, &iter,
			   TIMESTAMP_COLUMN, device_timestamp(rxo->timestamp),
			   TX_COLUMN, FALSE,
			   ERR_COLUMN, TRUE,
			   ARBID_COLUMN, ICS_NO_ARB_ID,
			   DESCRIPTION_COLUMN, "Rx Overflow",
			   BUS_COLUMN, rxo->netid,
			   -1);
}

static void
tx_complete_callback(struct ics_stream_state *ss, struct ics_tx_complete *txc,
		void *tempListStoreAndNumPackets)
{
	/*
	 * XXX: unimplemented
	 */

}

static void
can_error_callback(struct ics_stream_state *ss, struct ics_can_error *ce,
		void * user_data)
{
	GtkListStore * listStore = (GtkListStore *) user_data;
	GtkTreeIter iter;
	double timestamp;
	const gchar * errStr;
	gint offset = 0;

	/* Note: the order of these statements may result
	 * in less helpful error messages */
        if (ce->txwar)
		errStr = "Transmit Error Warning";
	else if (ce->rxwar)
		errStr = "Receive Error Warning";
	else if (ce->ewarn)
		errStr = "Error Warning";
	else if (ce->rx1ovr)
		errStr = "Receive Buffer 1 Overflow ";
	else if (ce->rx0ovr)
		errStr = "Receive Buffer 0 Overflow ";
	else if (ce->txbo)
		errStr = "Bus-Off Error";
	else if (ce->txep)
		errStr = "Transmit Error-Passive";
	else if (ce->rxep)
		errStr = "Receive Error-Passive";

	gtk_list_store_append(listStore, &iter);
	gtk_list_store_set(listStore, &iter,
			   TIMESTAMP_COLUMN, device_timestamp(ce->timestamp),
			   TX_COLUMN, FALSE,
			   ERR_COLUMN, TRUE,
			   ARBID_COLUMN, ICS_NO_ARB_ID,
			   DESCRIPTION_COLUMN, errStr,
			   BUS_COLUMN, ce->netid,
			   -1);

}

static void
pic_overflow_callback(struct ics_stream_state *ss, void * user_data)
{
	GtkListStore * listStore = (GtkListStore *) user_data;
	GtkTreeIter iter;

	gtk_list_store_append(listStore, &iter);
	gtk_list_store_set(listStore, &iter,
			   TX_COLUMN, FALSE,
			   ERR_COLUMN, TRUE,
			   ARBID_COLUMN, ICS_NO_ARB_ID,
			   DESCRIPTION_COLUMN, "PIC Overflow",
			   -1);
}

static void
main51_overflow_callback(struct ics_stream_state *ss,
		struct ics_main51_overflow *mo, void * user_data)
{
	GtkListStore * listStore = (GtkListStore *) user_data;
	GtkTreeIter iter;

	gtk_list_store_append(listStore, &iter);
	gtk_list_store_set(listStore, &iter,
			   TX_COLUMN, FALSE,
			   ERR_COLUMN, TRUE,
			   ARBID_COLUMN, ICS_NO_ARB_ID,
			   DESCRIPTION_COLUMN, "Main51 Overflow",
			   BUS_COLUMN, mo->netid,
			   -1);
}

/* This function should probably execute somewhat quickly */
static void 
can_callback(struct ics_stream_state *ss,
		 struct ics_can_msg *msg, void * user_data)
{
	GtkListStore * listStore = (GtkListStore *) user_data;
	GtkTreeIter iter;
	double timestamp;
	guint8 * data;

	/* 
	 * msg is on the stack of our caller so we have to make our own
	 * copy of anything we want to keep
	 */

	data = (guint8 *) g_malloc(ICS_MAX_CAN_DATA + 1); /* I sure hope this gets freed */
	memcpy(data, msg->data, msg->data_len);
	data[ICS_MAX_CAN_DATA] = msg->data_len; /* sigh - one byte is better than another column */

	gtk_list_store_append(listStore, &iter);
	gtk_list_store_set(listStore, &iter, 
			TIMESTAMP_COLUMN, device_timestamp(msg->timestamp),
			TX_COLUMN, FALSE,
			ERR_COLUMN, FALSE,
			ARBID_COLUMN, msg->can_id,
			DATA_BYTES_COLUMN, data,
			BUS_COLUMN, msg->netid,
			-1);
}

#define ERROR_LEN 256

/**
 * Take care of all the mess associated with opening up the neoVI/ValueCAN.
 * May or may not be sufficently generic for widespread use...see libintrepidcs.
 * @param deviceFileName Open this device.
 * @param listStore the structure in which to store incoming packets.
 * @param ifaceMon callback triggered when a packet is received - should call device_process_packet
 * and handle errors appropriately.
 * @param errStr Errors will be returned as dynamically allocated gchar * and must be g_free'd.
 * @return TRUE on successful initialization, FALSE on error.
 */
gboolean
device_init(const gchar * deviceFileName, GtkListStore * listStore, GIOFunc ifaceMon, gchar ** errStr)
{
	char error[ERROR_LEN];
	struct ics_packet p;

	iface = ics_interface_init(deviceFileName, error, ERROR_LEN);
	if (iface == NULL)
	{
		*errStr = g_strconcat("Couldn't open device: ", error, NULL);
		return FALSE;
	}

	ss = ics_stream_init(iface, error, ERROR_LEN);
	if(ss == NULL)
	{
		*errStr = g_strconcat("Couldn't create stream state: ", error, NULL);
		return FALSE;
	}

	ics_packet_create_can_disable(&p);
	if(!ics_interface_write_packet(iface, &p))
	{
		*errStr = g_strconcat("Couldn't write disable packet: ", error, NULL);
		return FALSE;
	}

	ics_packet_create_can_enable(&p);
	if(!ics_interface_write_packet(iface, &p))
	{
		*errStr = g_strconcat("Couldn't write enable packet: ", error, NULL);
		return FALSE;
	}

	iochan = g_io_channel_unix_new(ics_interface_fd(device_iface()));
	g_io_channel_set_close_on_unref(iochan, FALSE);
	g_io_channel_set_encoding(iochan, NULL, NULL);
	g_io_channel_set_buffered(iochan, FALSE);

	incomingPacketHandlerTag = g_io_add_watch(iochan,
						  G_IO_IN,
						  ifaceMon,
						  NULL); 

	ics_stream_set_device_status_callback(ss, dev_stat_callback, NULL);
	ics_stream_set_rx_overflow_callback(ss, rx_overflow_callback, listStore);
	ics_stream_set_tx_complete_callback(ss, tx_complete_callback, tempListStoreAndNumPackets);
	ics_stream_set_can_error_callback(ss, can_error_callback, listStore);
	ics_stream_set_pic_overflow_callback(ss, pic_overflow_callback, listStore);
	ics_stream_set_main51_overflow_callback(ss, main51_overflow_callback, listStore);
	ics_stream_set_hscan_callback(ss, can_callback, listStore);
	ics_stream_set_swcan_callback(ss, can_callback, listStore);
	ics_stream_set_mscan_callback(ss, can_callback, listStore);
	ics_stream_set_lsftcan_callback(ss, can_callback, listStore);
}

/**
 * Close the device.
 */
void
device_close(void)
{
	/* reset timestamp */
	numDevStatPackets = 0;
	
	if(iface)
	{
		ics_interface_release(iface);
		iface = NULL;
	}

	if(ss)
	{
		ics_stream_release(ss);
		ss = NULL;
	}

	if(!g_source_remove(incomingPacketHandlerTag))
	{
		g_warning("Couldn't remove source");
	}
	g_io_channel_unref(iochan);
	iochan = NULL;
}

/**
 * Get a new packet, check it for validity, and handle it appropriately.
 * @param error Pointer to gchar * to contain any errors, must be g_free'd.
 */
gboolean
device_process_packet(gchar ** error)
{
	int is_valid;
	struct ics_packet p;
	
	if(ics_interface_read_packet(iface, &p, &is_valid) == 0)
	{
		error = g_strconcat("Couldn't read packet: ", ics_interface_err_str(iface), NULL);
		return FALSE;
	}

	if(!is_valid)
	{
		error = g_strdup("Invalid packet");
		return FALSE;
	}

	ics_stream_process_packet(ss, &p);
	return TRUE;
}

double
device_timestamp(uint16_t timestamp)
{
	double correctedTimestamp;

	correctedTimestamp = timestamp * ics_stream_get_timestamp_tick(ss) +
				ics_stream_get_timestamp_overflow(ss) * numDevStatPackets;

	return correctedTimestamp;
}

/**
 * Exposes the wrapped ics_interface object for use in libintrepidcs calls.
 * @return handle to the ics_interface representing the physical device.
 */
struct ics_interface *
device_iface(void)
{
	return iface;
}
